/* return a remote connection's host name, or NULL if it */
/* can't be looked up */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <resolv.h>

#ifdef	DOUBLE_RES_LOOKUPS
struct hostent *double_reverse_check ();
#endif	/* DOUBLE_RES_LOOKUPS */

char *
getremotename(s)
int s;			/* connected socket fd */
{
    static char rname[256];
    struct hostent *host;
    struct sockaddr_in rs, sin;
    int rs_size = sizeof(rs);
    char *inet_ntoa();
    char tbuf[256], *tmp;

    /* Don't want to wast all our time in resolving stuff... */
    _res.retrans = 3;
    _res.retry = 2;
#ifdef VERBOSE
    sprintf(tbuf, "[INFO] retrans = %d, retry = %d, options = %d",
	    _res.retrans, _res.retry, _res.options);
    mdb(tbuf);
    mdb("[INFO] Starting hostname lookup");
#endif
    bzero(rname, 256);
    /* get address of remote user */
    if (getpeername(s, (struct sockaddr *)&rs, &rs_size) < 0) {
	perror("server: getpeername failed");
	return(NULL);
    }

#ifdef	DOUBLE_RES_LOOKUPS
    if ( (host = double_reverse_check (rs.sin_addr)) == (struct hostent *)NULL )
    {
	sprintf (tbuf, "[INFO] %s fails double reverse",
	    inet_ntoa (rs.sin_addr));
	mdb (tbuf);
    }
#else
    host = gethostbyaddr((char *)&rs.sin_addr.s_addr,
	    (int)sizeof(rs.sin_addr.s_addr), AF_INET);
#endif	/* DOUBLE_RES_LOOKUPS */

    /* get hostname from table */
    if (host == 0) {
	    sprintf(rname, "[%s]", inet_ntoa (rs.sin_addr));
	    sprintf(tbuf, "[INFO] Can not resolve %s", 
		    inet_ntoa(rs.sin_addr));
	    mdb(tbuf);
	    }
    else {
	    strncpy(rname, host->h_name, sizeof (rname)-1);
    }

#ifdef	BOSSANOVA_TWEAK
    /* ok ok, so i'm playing a joke on bossanova folks :) */
    {
	/* 209.60.53.41 = D13C3529, reversed = 29353CD1 */
	if ( rs.sin_addr.s_addr == 0x29353cd1 )
	{
	    sprintf (tbuf, "[INFO] Overriding %s with foo.bar", rname);
	    mdb (tbuf);
	    strcpy (rname, "foo.bar");
	}
    }
#endif	/* BOSSANOVA_TWEAK */

    return(rname);
}

char *
getlocalname(s)
int s;			/* connected socket fd */
{
    static char rname[256];
    struct hostent *host;
    struct sockaddr_in rs;
    int rs_size = sizeof(rs);
    char *inet_ntoa();
    char tbuf[256];

    /* Don't want to wast all our time in resolving stuff... */
    _res.retrans = 3;
    _res.retry = 2;
#ifdef VERBOSE
    sprintf(tbuf, "[INFO] retrans = %d, retry = %d, options = %d",
	    _res.retrans, _res.retry, _res.options);
    mdb(tbuf);
    mdb("[INFO] Starting hostname lookup");
#endif
    bzero(rname, 256);
    /* get address of local interface */
    if (getsockname(s, (struct sockaddr *)&rs, &rs_size) < 0) {
	perror("server: getsockname failed");
	return(NULL);
    }

    /* get hostname from table */
    if ((host = gethostbyaddr((char *)&rs.sin_addr.s_addr,
      (int) sizeof(rs.sin_addr.s_addr), AF_INET)) == 0)
    {
	sprintf(rname, "[%s]", inet_ntoa(rs.sin_addr));
	sprintf(tbuf, "[INFO] Can not resolve %s", 
	    inet_ntoa(rs.sin_addr));
	mdb(tbuf);
    }
    else
    {
	strcpy(rname, host->h_name);
    }
    return(rname);
}

#ifdef	DOUBLE_RES_LOOKUPS
struct hostent *
double_reverse_check (in)
    struct in_addr in;
{
    struct hostent *hp;
    static char *save_host = (char *)NULL;

    /* look up the ip address */
    hp = gethostbyaddr ((char *)&in.s_addr, sizeof (in.s_addr), AF_INET);
    if ( hp == (struct hostent *) NULL )
	return ( (struct hostent *)NULL );

    /* if we've been called before and have allocated memory, free it up */
    if ( save_host != (char *) NULL )
	free (save_host);

    /* save a copy of static memory which will be overwritten */
    if ( (save_host = (char *) strdup (hp->h_name)) == (char *) NULL )
	return ( (struct hostent *)NULL );

    /* now, look up the host */
    hp = gethostbyname (save_host);

    if ( hp == (struct hostent *)NULL )
	return ( (struct hostent *)NULL );

    while ( *hp->h_addr_list )
    {
	struct in_addr tin;
	bcopy (*hp->h_addr_list, &tin, hp->h_length);
	if ( tin.s_addr == in.s_addr ) 
	{
	    /* if hp->h_name isn't the correct hostname, fudge it */
	    if ( strcmp (hp->h_name, save_host) != 0 )
		hp->h_name = save_host;

	    return (hp);
	}
	(hp->h_addr_list)++;
    }

    return ( (struct hostent *)NULL );
}
#endif	/* DOUBLE_RES_LOOKUPS */
