/* Copyright (c) 1988 Carrick Sean Casey. All rights reserved. */

/* main program loop */
/* active clients, new connections, and perform asynch dungeon action */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#ifdef FCNTL
#include <fcntl.h>
#endif /* FCNTL */
#include "select.h"
#include "ipcf.h"
#include "externs.h"
#include "../server/server.h"

/* Stolen from Solaris land. I don't know why (unsigned) matters, but it does */
#ifdef sun
#ifdef FD_ISSET
#undef FD_ISSET
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & ((unsigned) 1 << ((n) % NFDBITS)))
#endif
#endif 

extern int errno;
extern int restart;
fd_set fdr, efdr;
int timeisup;


#ifdef SYSV 
void asignal ( int signo, void (*func)(int) ) {
  struct sigaction actung, foo;
  actung.sa_handler = func;
  sigemptyset(&actung.sa_mask);
  actung.sa_flags = SA_RESTART;
#ifdef SA_INTERRUPT  
  actung.sa_flags |= SA_INTERRUPT;
#endif
  sigaction ( signo, &actung, &foo );
  }
#endif

serverserve()
{
	int ret, i;
	void gotalarm();
	long loopcount = 0; /* used to break loop for periodic didpoll */
	char tbuf[256];

	/* look at fd where new clients connect */
	if (restart == 0) {
		FD_ZERO(&fdset);
		FD_ZERO(&ignorefds);
		FD_SET(port_fd, &fdset);
		}
	else 
		for (i = 0; i < MAX_USERS; i++)
		    cbufs[i].new = 1;

	if (port_fd > highestfd)
		highestfd = port_fd;

#ifdef SYSV
	asignal(SIGALRM, gotalarm);
#else
	signal(SIGALRM, gotalarm);
#endif
	timeisup = 0;

	for (;;) {
		if (polldelay) {
			if (setitimer(ITIMER_REAL, polldelay, 0) < 0) {
				perror("server: setitimer failed");
				exit(-1);
			}
			timeisup = 0;
			sigpause(0); /* WARNING: i *think* this should be on */
		}

		bcopy((char *) &fdset, (char *) &fdr,
		  (unsigned int) sizeof(fdset));
		bcopy((char *) &fdset, (char *) &efdr,
		  (unsigned int) sizeof(fdset));

		/* wait for input */
		/* NULL should be changed to a time value that reflects */
		/* how much activity is going on */

		setpolltimeout(BUSY_POLLWAIT);

		if ((ret = select(highestfd+1, &fdr, 0, &efdr, polltimeout)) < 1) {
			if (ret == 0 || errno == EINTR) {
				s_didpoll(0);
				timeisup = 0;
			} else {
				char buf[1024];
				sprintf (buf, "select: %s", strerror(errno));
				mdb(buf);
				/* perror("select"); */
				/* exit(-1); */
			}
		}

		if (ret > 0)
			sdoinput();
		if (timeisup || !(++loopcount % 10))
		{
#ifdef	JRL
		    int i;

		    /* for all users, check ignore/held state */
		    for (i = 0; i < MAX_USERS; i++)
		    {
			/* if it's currently being held */
			if (FD_ISSET(i, &ignorefds) > 0)
			{
			    /* and it's ok to read from it again */
			    if ( ok2read (i) == 1 )
				unignore (i);
			}
		    }
#endif	/* JRL */
		    s_didpoll(0);
		}
	}
}

void
gotalarm(n)
int n;
{
	timeisup++;
	signal(n, gotalarm);
}

sdoinput()
{
	int n;
	register int x;
	register unsigned long bits;

	for (x = 1; x <= highestfd; x++)
		if (FD_ISSET(x, &efdr) > 0) {
			disconnectuser(x);
			}

	/* examine set of file descriptors */
	for (x = 1; x <= highestfd; x++)
		if (FD_ISSET(x, &fdr) > 0) {
			if (x == port_fd) {
				/* new connect on advertised socket */
				if ((n = _newconnect(x)) > 0)
					/* let server init the user */
					s_new_user(n);
			} else {
				/* fd is client input */
#ifdef	JRL
				if ( ok2read (x) == 1 )
				{
#endif	/* JRL */
				    switch(_readpacket(x, &cbufs[x]))
				    {
				    case  1:
					    /* complete packet */
					    s_packet(x, cbufs[x].buf);
					    break;
				    case  0:
					    /* incomplete packet */
					    break;
				    case -1:
					    /* error */
					    break;
				    case -2:
					    /* close connection */
					    disconnectuser(x);
					    break;
				    }
#ifdef	JRL
				}
				else
				{
				    /* add to ignore and remove from fdset */
				    ignore (x);
				}
#endif	/* JRL */
			}
		}
}

