/* Copyright (c) 1991 by John Atwood deVries II. */
/* For copying and distribution information, see the file COPYING. */

#include "server.h"
#include "config.h"
#include "externs.h"
#include "groups.h"
#include "protocol.h"
#include "version.h"
#include <sys/wait.h>
#include <sys/file.h>
#include <signal.h>
#include <strings.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <fcntl.h>
#ifdef SYSV
/* need this for F_OK and friends */
#include <unistd.h>
#endif

main (argc, argv)
int argc;
char **argv;
{
	int c;
	int ret;
	extern char *optarg;
	int clearargsflg = 0;
	int forkflag = 0;
	int port = DEFAULTPORT;
	int i;
	long time();
	long TheTime;
	int pidf;
	char tbuf[255], *bindhost = (char *) NULL;
	struct rlimit rlp;

	/* save off args that we need in /restart */
	restart_argc = argc+2;	/* room for the -R and NULL args */
	restart_argv = (char **) malloc (restart_argc * sizeof (char *));

	if ( restart_argv == (char **)NULL )
	{
	    perror ("malloc(restart_args)");
	    exit (1);
	}

	for ( i = 0; i <= argc; i++ )
	    restart_argv[i] = (argv[i] == NULL) ? NULL : strdup (argv[i]);

	restart = 0;

	setbuf(stdout, (char *) 0);

	while ((c = getopt(argc, argv, "cp:fRb:")) != EOF) {

		switch (c) {

		case 'b':
			bindhost = optarg;
			break;

		case 'R':
			restart++;
			break;

		case 'f':
			forkflag++;
			break;

		case 'c':
			clearargsflg++;
			break;

		case 'p':
			port = atoi(optarg);
			break;

		case '?':
		default:
			puts("usage: icbd [-b host] [-p port] [-cRf]");
			puts("-c	 wipe args from command line");
			puts("-R	 restart mode");
			puts("-f	 don't fork");
			puts("-p port	 bind to port instead of default");
			puts("-b host	 bind socket to \"host\"");
			exit(-1);
			break;
		}
	}

#ifdef RLIMIT_NOFILE
   getrlimit(RLIMIT_NOFILE, &rlp);
   rlp.rlim_cur = MAX_USERS + 1;
   if (setrlimit(RLIMIT_NOFILE, &rlp) < 0) {
#ifdef SYSV
	perror("setrlimit (ignored, running server anyway)");
#else
	perror("setrlimit");
	exit(1);
#endif /* SYSV */
	}
#endif /* RLIMIT_NOFILE */

   if (restart == 0) {

	open_logs();

	/* get our hostname */
	if ((ret=gethostname(thishost,MAXHOSTNAMELEN)) < 0) {
		sprintf(tbuf, "[ERROR] gethostname failed: %s", 
			strerror(errno));
		mdb(tbuf);
		}

	if (strcmp(thishost, SHORT_HOSTNAME) == 0)
		strcpy(thishost, FQDN);

	/*
	 * here's the scoop with this code:
	 *
	 * 1- don't want to remove rich's special case SHORT/FQDN code yet
	 * 2- if BIND_HOSTNAME is set in config.h, bind to only that ip
	 * 3- if -b arg is used, override with its value
	 */

#ifdef	BIND_HOSTNAME
	/* if bindhost is null (unset), let BIND_HOSTNAME set it */
	if ( (char *)NULL == bindhost )
	{
	    bindhost = BIND_HOSTNAME;
	    strcpy (thishost, BIND_HOSTNAME);
	}
#endif	/* BIND_HOSTNAME */

	if ( (char *) NULL != bindhost )
	{
	    /*
	     * if an actual host has been listed, reset thishost to it,
	     * otherwise, we are mapping to all interfaces and using the
	     * existing thishost value for the name
	     */
	    if ( '\0' != *bindhost )
	    {
		(void) strcpy (thishost, bindhost);
	    }
	}

	if (makenewport(bindhost, port) < 0) {
		sprintf(tbuf, "[ERROR] makenewport failed: %s", 
			strerror(errno));
		mdb(tbuf);
		exit (-1);
	}

	if (forkflag == 0)
		if (fork()) exit(0);

	if (clearargsflg)
		clearargs(argc, argv);

	/* initialize out tables  */
	init_groups();
	clear_users();

	for (i = 0; i < MAX_REAL_USERS; i++)
	    S_kill[i] = 0;

	/* invent our resident automagical user */
	TheTime = time(NULL);
#ifdef	JRL
	fill_user_entry(NICKSERV, "server",thishost,
		"Server", "", "ICB", TheTime, 0, 1, PERM_NULL);
#else
	fill_user_entry(NICKSERV, "server",thishost,
		"Server", "", "ICB", TheTime, 0);
#endif	/* JRL */
	u_tab[NICKSERV].login = 1;
	u_tab[NICKSERV].t_on = TheTime;
	u_tab[NICKSERV].t_recv = TheTime;
	strcpy(u_tab[NICKSERV].realname, "registered");
	fill_group_entry(0, "ICB", "...here to serve you!", SUPERSECRET,
		RESTRICTED, NICKSERV, QUIET);
	nickwritetime(NICKSERV, 0);

	sprintf(mbuf, "[INFO] ICB revision %s on %s.", VERSION, thishost);
	mdb(mbuf);
	sprintf(mbuf, "[INFO] There are %d users possible.", MAX_USERS);
	mdb(mbuf);
	sprintf(mbuf, "[INFO] Of those, %d are real.", MAX_REAL_USERS);
	mdb(mbuf);
	
	setpolldelay(POLL_DELAY);
	setpolltimeout(IDLE_POLLWAIT);

	trapsignals();
   }
   else {
	extern int port_fd;
	extern char *getlocalname();
	char	*th = (char *)NULL;

	open_logs();
	/* get our hostname */
	if ((ret=gethostname(thishost,MAXHOSTNAMELEN)) < 0) {
		sprintf(tbuf, "[ERROR] gethostname failed: %s", 
			strerror(errno));
		mdb(tbuf);
		}

	if (strcmp(thishost, SHORT_HOSTNAME) == 0)
		strcpy(thishost, FQDN);

	setpolldelay(POLL_DELAY);
	setpolltimeout(IDLE_POLLWAIT);
	init_groups();
	clear_users();
	trapsignals(); 
	mdb("[RESTART] Reloading server data");
	icbload();

	if ( (th = getlocalname (port_fd)) != (char *) NULL )
	{
	    bzero(thishost, sizeof (thishost));
	    strncpy(thishost, th, sizeof (thishost));
	}

	for (i = 0; i < MAX_USERS; i++)
	    if (u_tab[i].login == 1)
		sendimport(i, "Restart", "Server is back up.");

    }

    if ((pidf = open(PID_FILE, O_WRONLY|O_CREAT|O_TRUNC)) >= 0)
    {
	    /* write the process id on line 1 */
	    sprintf(mbuf, "%d\n", getpid());
	    write(pidf, mbuf, strlen(mbuf));
	    fchmod (pidf, 0644);
	    close(pidf);
    }

#ifdef RLIMIT_NOFILE
    getrlimit(RLIMIT_NOFILE, &rlp);
    sprintf(tbuf, "[INFO] Max fds set to %d", rlp.rlim_cur);
    mdb(tbuf);
#endif

#ifdef	JRL
    for (i = 0; i < MAX_REAL_USERS; i++)
    {
	pong_req[i] = -1;
	timerclear (&ping_time[i]);
    }
#endif	/* JRL */

    /* start the serve loop */
    if (serverserve() < 0) {
	sprintf(tbuf, "[ERROR] serverserve failed: %s", 
		strerror(errno));
	mdb(tbuf);
	exit(-1);
    }

    /* all's well that ends */
    exit(0);
}

open_logs()
{
	/* used to have non-portable O_FSYNC */
	if((icbd_log = open(ICBDLOG,O_WRONLY|O_CREAT|O_APPEND, 0644)) < 0) {
		fprintf(stderr, "icbd: could not open log file.\n");
		perror("ICBDLOG open");
		exit(1);
		}
}

close_logs()
{
	close(icbd_log);
}

void
cycle_logs()
{
	(void) close_logs();
	(void) open_logs();
}

void
icbexit()
{
	int user;

	mdb("someone told us to exit");
	for (user=0; user < MAX_REAL_USERS; user++) {
		if (u_tab[user].login == 1) {
			disconnectuser(user);
		}
	}
	exit(0);
}
