/* Copyright (c) 1991 by John Atwood deVries II. */

/* handle various messages from the client */

#include "server.h"
#include "externs.h"
#include "lookup.h"
#include "groups.h"
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#ifdef	JRL
# include <sys/time.h>	/* for gettimeofday */
# ifndef	timersub
#  define timersub(tvp, uvp, vvp)					\
        do {								\
                (vvp)->tv_sec = (tvp)->tv_sec - (uvp)->tv_sec;		\
                (vvp)->tv_usec = (tvp)->tv_usec - (uvp)->tv_usec;	\
                if ((vvp)->tv_usec < 0) {				\
                        (vvp)->tv_sec--;				\
                        (vvp)->tv_usec += 1000000;			\
                }							\
        } while (0)
# endif		/* !timersub */
#endif	/* JRL */

/* open message */
/*
   note that we need to know who sent the message.  it doesn't
   come with the sender's nick, so we have to reformat the message
   before sending it to the group
*/

#ifdef	JRL
int u_recv_count[MAX_USERS];
#endif	/* JRL */

openmsg(n, pkt)
int n;
char *pkt;
{
	long TheTime;
	int gi;

        if (u_tab[n].login == 1) {
        	TheTime = time(NULL);

#ifdef	JRL
		if ( !strcmp (u_tab[n].group, "1") )
		{
		    u_recv_count[n]++;
		    if ( u_tab[n].t_recv == TheTime )
		    {
			if ( u_recv_count[n] > 2 )
			{
			    extern int is_booting;

			    /* tell 'em it happened. */
			    sendstatus(n, "Boot", "No spamming in group 1.");
			    sprintf(mbuf,"%s was auto-booted for spamming.",
				    u_tab[n].nickname);
			    s_status_group(1,1,n, "Boot", mbuf);
			    /* fake an s_change to group ICB */
			    strcpy (fields[1], "ICB");
			    is_booting = 1;
			    s_change(n, 2);
			    is_booting = 0;
			}
		    }
		    else
		    {
			u_recv_count[n] = 0;
		    }
		}
#endif	/* JRL */

	        /* record the time */
                u_tab[n].t_recv= TheTime;

		if (split(pkt) != 1) {
			mdb("[ERROR] got bad open message packet");
		} else {
#ifdef VERBOSE
		   sprintf(mbuf,"[OPEN] %d", n);
		   mdb(mbuf);
#endif
		   gi = find_group(u_tab[n].group);
		   if (g_tab[gi].volume == QUIET)
			senderror(n,
			"Open messages not permitted in quiet groups.");
#ifdef JRL
		    /*
		     * if the group is CONTROLLED,  and they
		     * aren't the moderator, check to make sure they are in
		     * in the talk list
		     */
		    else if ( (g_tab[gi].control == CONTROLLED
			&& g_tab[gi].mod != n)
		        && ( (nlpresent (u_tab[n].nickname,
			    *g_tab[gi].n_talk) == 0)
			&& (strlen (u_tab[n].realname) <= 0
			    || nlpresent (u_tab[n].nickname,
				*g_tab[gi].nr_talk) == 0) ) )
		    {
			senderror (n,
	    "You do not have permission to send open messages in this group.");
			sprintf (mbuf,
			    "Use \"/m %s may i talk?\" to request permission.",
			    u_tab[g_tab[gi].mod].nickname);
			sendstatus (n, "INFO", mbuf);
		    }
#endif /* JRL */
		   else if (count_users_in_group(g_tab[gi].name) < 2)
			senderror(n,
			"No one else in group!");
		   else {
			s_send_group(n);
		   }
		}
	} else {
		mdb("[ERROR] cannot send open messages until logged in");
	}
}


/*
   they sent us a login message. we need to put their info into 
   the user information table and then send them back a loginok message.
   for any field they didn't specify, come up with a plausible default.
   we also need to handle "w" vs. "login"
   we also need to dump them if they try using a:
	nickname already in use -or-
	group they aren't allowed to enter.
   their loginid may NOT be blank or null
*/

loginmsg(n, pkt)
int n;		/* socket on which they sent message */
char *pkt;	/* packet buffer */
{
    int num_fields;
    char which_group[MAX_NICKLEN+4];
    char temp[MAX_NICKLEN+4];
    char one[255], two[255], three[255];
    int len;
    int ret;
    int how_many;
    int i, j;
    long TheTime;
    int gi;
    int target_user;
    char * cp;
    extern char * getremotename();
    int access_file;
    char c;
#ifdef	JRL
    long perms = PERM_NULL;
#endif	/* JRL */

    if (u_tab[n].login == 1)
    {
	senderror(n, "Already logged in.");
	return 1;
    }

    u_tab[n].login = 0;
    /* just in case */
    TheTime = time(NULL);

    num_fields = split(pkt);

    if (num_fields < 4) return -1;

    len = strlen(fields[2]);

    bzero(temp, MAX_NICKLEN+4);
    bzero(which_group, MAX_NICKLEN+4);

    if (len == 0) /* no group specified, give them one */
    {
	strcpy(which_group,"1");
    }
    else
    {
	how_many = (len > (MAX_NICKLEN+3))? (MAX_NICKLEN+3): len;
	strncpy(which_group, fields[2], how_many);
    }

    if ( strcasecmp(fields[3],"w") == 0)
    {
	/* who only */
	fields[0][0] = 'w';
	fields[0][1] = '\000';
	fields[1][0] = '\000';

	s_who(n, 2);
	return -1; /* fail */
    }
    /*
     * this is basically the same as "w" above, but properly supports
     * passing a group name. some old clients erroneously default the
     * group to "1" which makes "icb -w" only list group 1. yes, it's
     * wrong, but we want to keep people happy
     */
    else if ( strcasecmp(fields[3],"wg") == 0)
    {
	/* who only */
	fields[0][0] = 'w';
	fields[0][1] = '\000';

	/* if we got a groupname, use it */
	if ( num_fields >= 3 && fields[2][0] != '\0' )
	    sprintf (fields[1], "%s\000", fields[2]);
	else
	    fields[1][0] = '\000';

	s_who(n, 2);
	return -1; /* fail */
    }
    else if ( !strcasecmp (fields[3], "login") )
    {
	/* regular login */

	/* make sure the nickname hasn't already been taken */
	if (find_user(fields[1]) >= 0)
	{
	    /* oops, someone already has this nick */
	    senderror(n,"Nickname already in use.");
	    return(-1);
	}

	/* make sure it isn't a null nickname */
	if ((strlen(fields[1]) <= 0) || (strlen(fields[1]) > MAX_NICKLEN))
	{
	    sprintf(mbuf, "Nickname must be between 1 and %d characters.",
		    MAX_NICKLEN);
	    senderror(n, mbuf);
	    return -1;
	}

	/* make sure the length of the loginid isn't wrong */
	if ((strlen(fields[0]) <= 0) || (strlen(fields[0]) > MAX_IDLEN))
	{
	    sprintf(mbuf, "Login id must be between 1 and %d characters.",
		MAX_IDLEN);
	    senderror(n, mbuf);
	    return -1;
	}

	for (i = 0; i < strlen(fields[0]); i++)
	    if (isalnum(fields[0][i]) == 0)
	    {
		senderror(n,
		    "Login id must contain only alphanumeric characters.");
		return -1;
	    }

	/* This is set up in s_new_user now */
	cp = u_tab[n].nodeid;

#ifdef	JRL
	if ( (perms = get_perms (n, fields[0])) < 0L )
	{
	    return perms;
	}
	else
	{
	    if ( perms & PERM_DENY )
	    {
		sprintf (mbuf, "Login denied.");
		senderror(n, mbuf);
		mdb (mbuf);
		sprintf (mbuf, "Email %s for details.", SERVER_ADMIN);
		senderror(n, mbuf);
		return (-1);
	    }

	    if ( perms != PERM_NULL )
	    {
		mbuf[0] = '\0';
		if ( perms & PERM_SLOWMSGS ) { strcat (mbuf, "slowmsgs"); }
		sendstatus(n, "Permissions", mbuf);
	    }
	}
#endif	/* JRL */

	bzero(one, 255);
	bzero(three, 255);
	sprintf(one, "%s@%s", fields[0], cp);
	ucaseit(one);
	access_file = open(ACCESS_FILE, O_RDONLY);
	if (access_file >= 0)
	{
	    while ((i = read(access_file, &c, 1)) > 0)
	    {
		if (isspace(c))
		{
		    ucaseit(three);
		    if (wildmat(one, three))
		    {
			sprintf (mbuf, "Login refused for %s", one);
			senderror(n, mbuf);
			mdb (mbuf);
			bzero(two, 255);
			while ((i = read(access_file, &c, 1)) > 0)
			{
			    if (c == '\012') break;
			    strncat(two, &c, 1);
			}
			sprintf (mbuf, "Reason: %s", two);
			senderror(n, mbuf);
			sprintf (mbuf, "Email %s for details.", SERVER_ADMIN);
			senderror(n, mbuf);
			if (close(access_file) != 0)
			{
			    sprintf(mbuf, "[ERROR] Access File Close: %s",
				strerror(errno));
			    mdb(mbuf);
			}
			return -1;
		    }
		    bzero(three, 255);
		}
		else
		{
		    strncat(three, &c, 1);
		}
	    }

	    if (close(access_file) != 0)
	    {
		sprintf(mbuf, "[ERROR] Access File Close: %s",
		    strerror(errno));
		mdb(mbuf);
	    }
	}
	else
	{
	    sprintf(mbuf, "[ERROR] Access File Open: %s", strerror(errno));
	    mdb(mbuf);
	}
		
	/* get rid of nasty characters in the nickname */
	/* mdb("filtering nickname"); */
	filternickname(fields[1]);

	/* fill in what we can */
	/* mdb("filling in user entry"); */
#ifdef	JRL
	fill_user_entry(n, fields[0], cp, fields[1],
	    fields[4], "", 0, 0, 0, perms);
#else
	fill_user_entry(n, fields[0], cp, fields[1],
	    fields[4], "", 0, 0);
#endif	/* JRL */

	sprintf(mbuf, "[LOGIN] %d: %s@%s", n, fields[0], cp);
	mdb(mbuf);

	/* make sure they are allowed to use this nickname */
	if (strcasecmp(fields[1],"admin") == 0)
	{
	    if (!check_auth(n))
	    {
		/* oops, password is wrong */
		senderror(n,"Nickname already in use.");
		return(-1);
	    }
	}

	send_loginok(n);

	s_motd(n,2);

	if (which_group[0] == '@')
	{
	    if ((target_user = find_user(&which_group[1])) < 0)
	    {
		senderror (n,
		    "User not found or is in a secret or invisible group.");
		return(-1);
	    }
	    else
	    {
		int	g = find_group(u_tab[target_user].group);

		if ( ((g_tab[g].visibility == SUPERSECRET)
		    || (g_tab[g].visibility == SECRET))
		    && (strncmp("ADMIN", u_tab[n].nickname, MAX_NICKLEN)))
		{
		    senderror (n,
			"User not found or is in a secret or invisible group.");
		    return (-1);
		}
		else
		    strcpy (which_group, u_tab[target_user].group);
	    }
	}

#ifdef	OLD
	/* fake a s_change */
	strcpy(fields[1], which_group);
	/* mdb(fields[1]); */
	if (s_change(n,2) < 0)
	{
	    /* login fails because can't get into that group */
	    return(-1);
	}
#endif	/* OLD */

	/* could find group slot, all is ok */
	/* fill in last members of table entry */
	/* strcpy(u_tab[n].group, which_group); */
	u_tab[n].login = 1;
	u_tab[n].t_on = TheTime;
	u_tab[n].t_recv = TheTime;
	u_tab[n].t_sent = TheTime;
	u_tab[n].pri_n_hushed = (NAMLIST *) malloc(sizeof(NAMLIST));
	nlinit(u_tab[n].pri_n_hushed, MAX_HUSHED);
	u_tab[n].pub_n_hushed = (NAMLIST *) malloc(sizeof(NAMLIST));
	nlinit(u_tab[n].pub_n_hushed, MAX_HUSHED);
	u_tab[n].pri_s_hushed = (NAMLIST *) malloc(sizeof(NAMLIST));
	nlinit(u_tab[n].pri_s_hushed, MAX_HUSHED);
	u_tab[n].pub_s_hushed = (NAMLIST *) malloc(sizeof(NAMLIST));
	nlinit(u_tab[n].pub_s_hushed, MAX_HUSHED);
	u_tab[n].n_notifies = (NAMLIST *) malloc(sizeof(NAMLIST));
	nlinit(u_tab[n].n_notifies, MAX_NOTIFIES);
	u_tab[n].s_notifies = (NAMLIST *) malloc(sizeof(NAMLIST));
	nlinit(u_tab[n].s_notifies, MAX_NOTIFIES);

	sprintf(mbuf, "Welcome to ICB %s", u_tab[n].nickname);
	sends_cmdout(n, mbuf);

#ifdef	OLD
	/* this is moved below, along with the group initialization */
	bzero(one, 255);
	bzero(two, 255);
	bzero(three, 255);
	sprintf(one, "%s@%s", u_tab[n].loginid, u_tab[n].nodeid);
	ucaseit(one);
	strcpy(three, u_tab[n].nickname);
	ucaseit(three);
	for (j = 0; j < MAX_USERS; j++)
	    if ((n != j) && (u_tab[j].login > 0))
	    {
		if ((nlmatch(three, *u_tab[j].n_notifies) ||
		   nlmatch(one, *u_tab[j].s_notifies)) && 
		   (g_tab[find_group(u_tab[n].group)].visibility !=
		      SUPERSECRET))
		{
		    sprintf (two, "%s (%s) has just signed on",
			u_tab[n].nickname, one);
		    sendstatus(j, "Notify-On", two);
		}
	    }
#endif	/* OLD */

	/* check to see if we know this person */
	ret = nicklookup(-n, u_tab[n].nickname);

	/* we know this person but they're not validated */
	if (ret == -2)
	{
	    /*
	     * if password has a value in it, it's been passed from the
	     * client with the login message. if we aren't authenticated
	     * yet, try to.
	     */
	    if ( u_tab[n].password[0] != '\0' )
	    {
		if ( nickwrite (n, u_tab[n].password, 1) == 0 )
		{
		    ret = 0;
		}
	    }

	    if ( ret != 0 )
	    {
		sendstatus(n, "Register",
		    "Send password to authenticate your nickname.");
		j = 0;

		for (i = 0; i < MAX_GROUPS; i++)
		    if (strcasecmp(u_tab[n].nickname, g_tab[i].missingmod) == 0)
		    {
			sprintf(mbuf, "You are moderator of group %s",
			    g_tab[i].name);
			sendstatus(n, "Mod", mbuf);
			j++;
		    }

		if (j == 1)
		    sendstatus(n, "Mod",
    "You must register using /p <password> to regain mod of the above group.");

		if (j > 1)
		    sendstatus(n, "Mod",
    "You must register using /p <password> to regain mod of the above groups.");
	    }
	}

	if (ret == 0)
	{
	    /* we know this person */
	    if ( strcmp (u_tab[n].realname, "registered") )
	    {
		strcpy(u_tab[n].realname, "registered");
		sends_cmdout(n, "Nick registered");
		nickwritetime(n, 0);

		for (i = 1; i < MAX_GROUPS; i++)
		    if ((g_tab[i].modtimeout > 0.0) &&
			(strcmp(g_tab[i].missingmod, u_tab[n].nickname)==0))
			{
			    g_tab[i].modtimeout = 0;
			    g_tab[i].mod = n;
			    bzero(g_tab[i].missingmod, MAX_NICKLEN);
			    sprintf(mbuf, "%s is the active moderator again.",
				u_tab[n].nickname);
			    for (j = 1; j < MAX_REAL_USERS; j++)
				if ((strcasecmp(u_tab[j].group, g_tab[i].name) == 0)
				    && (j != n))
				    sendstatus(j, "Mod", mbuf);
			    sprintf(mbuf, "You are the moderator of group %s",
				g_tab[i].name);
			    sendstatus(n, "Mod", mbuf);
			}

		if ((i = nickckmsg(n)) > 0)
		    if (i == 1)
			sendstatus(n, "Message", "You have 1 message");
		    else
		    {
			sprintf(mbuf, "You have %d messages", i);
			sendstatus(n, "Message", mbuf);
		    }
	    }
	}
	else if ( ret != -2 )
	{
	    sendstatus(n, "No-Pass", "Your nickname does not have a password.");
	    sendstatus(n, "No-Pass", "For help type /m server ?");
	}

	bzero(one, 255);
	sprintf(one, "%s@%s", u_tab[n].loginid, u_tab[n].nodeid);
	ucaseit(one);
	for (i = 0; i < MAX_GROUPS; i++)
	{
	    if (
	       (nlpresent(u_tab[n].nickname, *g_tab[i].n_invites) > 0) ||
	       (nlmatch(one, *g_tab[i].s_invites)) ||
	       (nlpresent(u_tab[n].nickname, *g_tab[i].nr_invites) &&
		  (strlen(u_tab[n].realname) > 0)) ||
	       (nlmatch(one, *g_tab[i].sr_invites) && 
		  (strlen(u_tab[n].realname) > 0)))
	    {
		sprintf (mbuf, "Invited to: %s", g_tab[i].name);
		sends_cmdout(n, mbuf);
	    }
	}
		
#ifdef	JRL
	/*
	 * this is also done in nickwrite(), so only run it again
	 * if that did not run or did not succeed
	 */
	if ( strcmp (u_tab[n].realname, "registered") )
	{
	    for (i = 0; i < MAX_GROUPS; i++)
		talk_report (n, i);
	}
#endif	/* JRL */

	/* fake a s_change */
	strcpy(fields[1], which_group);
	/* mdb(fields[1]); */
	if (s_change(n,3) < 0)
	{
	    /* login fails because can't get into that group */
	    return(-1);
	}

	bzero(one, 255);
	bzero(two, 255);
	bzero(three, 255);
	sprintf(one, "%s@%s", u_tab[n].loginid, u_tab[n].nodeid);
	ucaseit(one);
	strcpy(three, u_tab[n].nickname);
	ucaseit(three);
	for (j = 0; j < MAX_USERS; j++)
	    if ((n != j) && (u_tab[j].login > 0))
	    {
		if ((nlmatch(three, *u_tab[j].n_notifies) ||
		   nlmatch(one, *u_tab[j].s_notifies)) && 
		   (g_tab[find_group(u_tab[n].group)].visibility !=
		      SUPERSECRET))
		{
		    sprintf (two, "%s (%s) has just signed on",
			u_tab[n].nickname, one);
		    sendstatus(j, "Notify-On", two);
		}
	    }

	return 0;
    }
    else
    {
	sprintf (mbuf, "Unknown login type \"%s\"", fields[3]);
	senderror (n, mbuf);
	return -1;
    }
}

/* command message */

cmdmsg(n, pkt)
int n;
char *pkt;
{
	int argc;
	long TheTime;

        if (u_tab[n].login == 1) {

                /* record the time */
                TheTime = time(NULL);
                u_tab[n].t_recv= TheTime;

		argc = split(pkt);

#ifdef VERBOSE
		if (strcmp(fields[0], "m") == 0)
		  sprintf(mbuf, "[COMMAND] %d: %s %s", n, fields[0],
			getword(fields[1]));
		else
		  sprintf(mbuf, "[COMMAND] %d: %s %s", n, fields[0], fields[1]);
		mdb(mbuf);
#endif
	
		switch(lookup(fields[0], command_table)) {

			case CMD_DROP:
				s_drop(n, argc);
				break;

			case CMD_RESTART:
				s_restart(n, argc);
				break;
			
			case CMD_SHUTDOWN:
				s_shutdown(n, argc);
				break;

			case CMD_WALL:
				s_wall(n, argc);
				break;

		 	case CMD_BEEP:
				s_beep(n, argc);
				break;

			case CMD_CANCEL:
				s_cancel(n, argc);
				break;

			case CMD_G:	
				s_change(n, argc);
				break;

			case CMD_INVITE:
				s_invite(n, argc);
				break;

			case CMD_PASS:
				s_pass(n, argc);
				break;

			case CMD_BOOT:
				s_boot(n, argc);
				break;

			case CMD_STATUS:
				s_status(n, argc);
				break;

			case CMD_TOPIC:
				s_topic(n, argc);
				break;

			case CMD_MOTD:
				s_motd(n, argc);
				break;

			case CMD_M:
				s_personal(n, argc);
				break;

			case CMD_ECHOBACK:
				s_echoback(n, argc);
				break;

			case CMD_NAME:
				s_name(n, argc);
				break;

			case CMD_V:
				s_version(n, argc);
				break;

			case CMD_W:
				s_who(n, argc);
				break;

			case CMD_INFO:
				s_info(n, argc);
				break;

			case CMD_NEWS:
				s_news(n, argc);
				break;

			case CMD_HUSH:
			case CMD_SHUSH:
				s_hush(n, argc);
				break;

			case CMD_HELP:
				s_help(n, argc);
				break;

			case CMD_EXCLUDE:
				s_exclude(n, argc);
				break;

			case CMD_SHUTTIME:
				s_shuttime(n, argc);
				break;

			case CMD_NOTIFY:
				s_notify(n, argc);
				break;

#ifdef	JRL
			case CMD_PING:
				s_ping(n, argc);
				break;

			case CMD_TALK:
				s_talk(n, argc);
				break;

			case CMD_NOBEEP:
				s_nobeep(n, argc);
				break;
#endif	/* JRL */

		default:
			if (lookup(fields[0], auto_table) >= 0)
				autoCommand(n, 0);
			else {
				sendstatus(n, "Server", "Unknown command");
				sprintf(mbuf,
			 		"[INFO] %d: Invalid command \"%s\"",
					n, fields[0]);
				mdb(mbuf);
				}
		}
	} else {
		mdb("[ERROR] cannot issue command until logged in");
	}
}


/* pong message */

pong(n, pkt)
int n;
char *pkt;
{
	long TheTime;
	char tbuf[255];

        if (u_tab[n].login == 1) {
#ifdef	JRL
		struct timeval now;
		gettimeofday (&now, NULL);
		if ( pong_req[n] != -1 )
		{
		    struct timeval diff;

		    timersub (&now, &ping_time[n], &diff);
		    sprintf (tbuf, "%s client responded in %d.%02d seconds.",
			u_tab[n].nickname, diff.tv_sec, diff.tv_usec / 10000);
		    sendstatus (pong_req[n], "PONG", tbuf);
		    pong_req[n] = -1;
		    timerclear (&ping_time[n]);
		}
#else	/* JRL */
		sprintf(tbuf, "[INFO] %d: pong message", n);
        	mdb(tbuf);

                /* record the time */
                TheTime = time(NULL);
                u_tab[n].t_recv= TheTime;
#endif	/* JRL */
	} else {
		 mdb("[ERROR] cannot send pong message until signed on");
	}
}

#ifdef	JRL

#include <netinet/in.h>

long
get_perms (n, login)
    int n;
    char *login;
{
    struct sockaddr_in	rs;
    int			rs_size = sizeof(rs);
    static FILE		*perms_fp = (FILE *) NULL;
    char		pbuf[BUFSIZ];
    long		perm = PERM_NULL;
    extern char		*strtok();

    /* until i can figure out why this is still leaving fd's open, we
     * return immediately since this isn't being used right now anyways
    return (PERM_NULL);
     */

    if (getpeername(n, &rs, &rs_size) < 0)
    {
	sprintf (mbuf, "getpeername() failed: %s", strerror(errno));
	senderror(n, mbuf);
	mdb (mbuf);
	return (-1L);
    }

    /*
     * this really shouldn't be necessary, but for some reason
     * there is some strange case where this is being left open, perhaps
     * from a timer interrupting this function's ability to complete
     */
    if ( perms_fp != (FILE *) NULL )
    {
	fclose (perms_fp);
	perms_fp = (FILE *)NULL;
    }

    if ( (perms_fp = fopen (PERM_FILE, "r")) == (FILE *)NULL )
    {
	sprintf(mbuf, "[ERROR] Permissions File Open: %s", strerror(errno));
	mdb(mbuf);
	return (PERM_NULL);
    }

    while ( fgets (pbuf, BUFSIZ, perms_fp) != (char *) NULL )
    {
	char	*type,
		*arg1,
		*arg2,
		*perms,
		*cp;

	if ( index ("\n\t#", pbuf[0]) != (char *) NULL )
	{
	    continue;
	}

	type = strtok (pbuf, " \t");
	if ( (arg1 = strtok ((char *)NULL, " \t")) == (char *)NULL )
	    { continue; }
	if ( (arg2 = strtok ((char *)NULL, " \t")) == (char *)NULL )
	    { continue; }
	if ( (perms = strtok ((char *)NULL, " \t\n")) == (char *)NULL )
	    { continue; }

	if ( !strcasecmp (type, "user") )
	{
	    if ( !strcasecmp (arg1, login) &&
		!strcasecmp (arg2, u_tab[n].nodeid) )
	    {
		perm = perm2val (perms);
		break;
	    }
	}
	else if ( !strcasecmp (type, "mask") )
	{
	    unsigned long addr, mask;

	    addr = inet_addr(arg1);
	    mask = inet_addr(arg2);

	    if ( (rs.sin_addr.s_addr & mask) == addr )
	    {
		perm = perm2val (perms);
		break;
	    }
	}
    }

    fclose (perms_fp);
    perms_fp = (FILE *)NULL;

    return (perm);
}

long
perm2val (str)
    char	*str;
{
    long	perm = PERM_NULL;
    char	*cp;

    cp = strtok (str, "|");
    do
    {
	if ( !strcasecmp (cp, "perm_null") ) { perm |= PERM_NULL; }
	else if ( !strcasecmp (cp, "perm_deny") ) { perm |= PERM_DENY; }
	else if ( !strcasecmp (cp, "perm_slowmsgs") ) { perm |= PERM_SLOWMSGS; }
    } while ( (cp = strtok ((char *) NULL, "|")) != (char *)NULL );

    return (perm);
}

/*
 * ok2read()
 * - this function is called every time a socket is ready to read some data.
 *   therefore, here is where we apply the SLOWMSGS permissions so that
 *   certain connections' data is read less frequently than normal
 */
ok2read (n)
    int n;
{
    if ( u_tab[n].perms & PERM_SLOWMSGS )
    {
	time_t TheTime;

	TheTime = time(NULL);

	/*
	 * if the current time - the last time they sent something
	 * is less than 1 second, ignore 'em
	 */
	if ( (TheTime - u_tab[n].t_recv) < 1 )
	{
	    /* printf ("ok2read(%d) last sent %d secs ago-- waiting\n", n, TheTime - u_tab[n].t_recv); /* DeBuG */
	    return (0);
	}

	/* printf ("ok2read(%d) is ok to read\n", n); /* DeBuG */

	return (1);
    }

    return (1);
}

#endif	/* JRL */
